/*
 * Virtual mouse driver for Linux
 * 
 * Skeleton based on:
 *  $Id: sermouse.c,v 1.17 2002/03/13 10:03:43 vojtech Exp $
 *
 *  Copyright (c) 1999-2001 Vojtech Pavlik
 *
 * Copyright (c) 2004 Dan Aloni
 */

#include <linux/module.h>
#include <linux/slab.h>
#include <linux/interrupt.h>
#include <linux/input.h>
#include <linux/config.h>
#include <linux/serio.h>
#include <linux/init.h>

MODULE_AUTHOR("Dan Aloni <da-x@colinux.org>");
MODULE_DESCRIPTION("Virtual mouse driver");
MODULE_LICENSE("GPL");

/*
 * comouse_interrupt() handles incoming characters, either gathering them into
 * packets or passing them to the command routine as command output.
 */

static irqreturn_t comouse_interrupt(struct serio *serio,
				     unsigned char data, unsigned int flags, struct pt_regs *regs)
{
	return IRQ_HANDLED;
}

/*
 * comouse_disconnect() cleans up after we don't want talk
 * to the mouse anymore.
 */

static void comouse_disconnect(struct serio *serio)
{
}

/*
 * comouse_connect() is a callback form the serio module when
 * an unhandled serio port is found.
 */

static void comouse_connect(struct serio *serio, struct serio_driver *dev)
{
}

static struct serio_driver comouse_dev = {
	.interrupt =	comouse_interrupt,
	.connect =	comouse_connect,
	.disconnect =	comouse_disconnect,
	.driver		= {
		.name	= "comouse",
	},
};

int __init comouse_init(void)
{
	serio_register_driver(&comouse_dev); 
	return 0;
}

void __exit comouse_exit(void)
{
	serio_unregister_driver(&comouse_dev); 
}

module_init(comouse_init);
module_exit(comouse_exit);
